/* sane - Scanner Access Now Easy.

   Copyright (C) 2002 Frank Zago (sane at zago dot net)
   
   This file is part of the SANE package.
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.
   
   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.
   
   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.
   
   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.
   
   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.
   
   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice. 
*/

/*
   $Id: teco3.c,v 1.8 2005-09-19 19:57:48 fzago-guest Exp $
   VM3552 (and maybe VM4552)
   VM6552 in progress
*/

/*--------------------------------------------------------------------------*/

#define BUILD 2			/* 2008/04/01 */
#define BACKEND_NAME teco3

/*--------------------------------------------------------------------------*/

#include "../include/sane/config.h"	/* see backend-writing.txt */

#include <ctype.h> /* isprint() */
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "../include/sane/sane.h"
#include "../include/sane/sanei.h"
#include "../include/sane/saneopts.h"
#include "../include/sane/sanei_scsi.h"
#include "../include/sane/sanei_debug.h"
#include "../include/sane/sanei_backend.h"
#include "../include/sane/sanei_config.h"
#include "../include/lassert.h"

#define TECO3_TESTCODE	/* compile more code for creating test images */
#include "teco3.h"

/*--------------------------------------------------------------------------*/

char const TECO_CONFIG_FILE[] = "teco3.conf";

#ifdef TECO3_TESTCODE

/* The config file may contain a line like "#TEST=xxx". 
 * This is *only* for debugging. It will `mess up' the image scanned.
 * Hint: scan the flag of Mauritius (landscape) from your "grandpa's wiki :)
 */
SANE_String_Const const test_mode_list[] = {
  "normal",	/* no test (reset to default behavior) */
  "RGB",	/* Write "RRRR GGGG BBBB" lines, with binary raster number.
		 * This produces three mostly grey pictures side by side.
		 * All lines regarded as red are at the left, blue is right.
		 * The first 6 pixels are overwritten with the line number.
		 */
  "bulk",	/* write "r000 r001 r002" (handled like correct data).
		 * Same treatment as grey data. The first pixel of a bunch
		 * shows the color of the previous line (adjacent to its left).
		 * This produces a similar trifold grey picture as "RGB",
		 * but colors are scrambled, and lines don't 'line up'.
		 */
  "raw",	/* Write "r000 r001 r002" (as recieved), with color pixel.
		 * Same as "bulk", but lines are read one by one, so you can
		 * see the color of each line as the first (leftmost) pixel.
		 */
  NULL
};

#endif

/*--------------------------------------------------------------------------*/

/* For scsi inquiry: All scanners with such a string are accepted: */
static SANE_String_Const const scanner_list[] = {
  "TECO VM3552",
  "TECO VM4552",	/* totally untested! */
  "TECO VM6552",
  NULL
};

/*--------------------------------------------------------------------------*/

static Teco_Scanner *scanners = NULL;	    /* List of attached scanners. */
static const SANE_Device **devlist = NULL;  /* for sane_get_devices */

/*--------------------------------------------------------------------------*/

/* Possible scan modes and their associated parameters: */

static SANE_String_Const const scan_mode_list[TECO_N_MODES+1] = {
  BLACK_WHITE_STR,
  GRAY_STR,
  COLOR_STR,
  NULL
/*
TODO: use predefined
	SANE_VALUE_SCAN_MODE_LINEART,
	SANE_VALUE_SCAN_MODE_GRAY,
	SANE_VALUE_SCAN_MODE_COLOR,
*/
};

static SANE_Frame const mode_frame[TECO_N_MODES] = { 
	SANE_FRAME_GRAY,
	SANE_FRAME_GRAY,
	SANE_FRAME_RGB,
};

static char const mode_4scsi[TECO_N_MODES] = { 0, 2, 5 }; /* for set_window() */
static char const mode_color[TECO_N_MODES] = { 1, 1, 3 }; /* #colors */
static char const mode_depth[TECO_N_MODES] = { 1, 8, 8 }; /* bits/color */


/*--------------------------------------------------------------------------*/

/* Gamma range */
static const SANE_Range gamma_range = {
  0,				/* minimum */
  255,				/* maximum */
  0				/* quantization */
};

/*--------------------------------------------------------------------------*/

/* List of dithering options. */
static SANE_String_Const const dither_list[] = {
  "Line art",
  "2x2",
  "3x3",
  "4x4 bayer",
  "4x4 smooth",
  "8x8 bayer",
  "8x8 smooth",
  "8x8 horizontal",
  "8x8 vertical",
  NULL
};

/*--------------------------------------------------------------------------*/

static const SANE_Range threshold_range = {
  0,				/* minimum */
  255,				/* maximum */
  0				/* quantization */
};


/******************************************************************************
 *
 * Local functions. These might be reused in other backends.
 */

#define MIN( a, b )	( (a)>(b) ? (b) : (a) )
#define MAX( a, b )	( (a)<(b) ? (b) : (a) )

/* Display a buffer in the log. */
static void
hexdump( int level, char const *title, unsigned char const *buf, int len )
{
  int duplicates = 0; /* grows 16+1 per identical line */
  int i;

  DBG( level, "%s: (%d bytes)\n", title, len );

  for ( i = 0; i < len; i += 16 )
  {
    char line_hex[3*16+1];
    char line_asc[16]; /* no \0 needed */
    int j;

    /* write the current line's values into the char arrays: */
    for ( j = 0; j < 16; j++ )
    {
      if ( i+j < len )
      {
	if ( i>0 && buf[i+j] == buf[i+j-16] )
	  duplicates++;
	sprintf (line_hex+3*j, "%3.2x", buf[i+j]);
	line_asc[j] = isascii(buf[i+j]) && isprint(buf[i+j]) ? buf[i+j] : '.';
      }
      else
      {
	line_hex[3*j] = line_asc[j] = '\0';
      }
    }

    /* print the char arrays, or detect another identical line: */
    if ( duplicates % (16+1) == 16 ) /* i.e. 16x incremented */
    {
      duplicates++; /* now 17, 34, ... */
    }
    else
    {
      if ( duplicates > 16 )
	DBG( level, "... (repeated %d times)\n", duplicates / (16+1) );

      duplicates = 0;
      DBG( level, "%.2x:%-48.48s   %-.16s\n", i, line_hex, line_asc );
    }
  }

  if ( duplicates ) /* identical lines down to the last line */
    DBG( level, "... (repeated %d times)\n", duplicates / (16+1) );
}


/* Returns the length of the longest string, including the terminating \0. */
static size_t
max_string_size ( SANE_String_Const const strings[] )
{
  size_t max_len = 0;

  for ( ; *strings; strings++ )
  {
    size_t const l = strlen( *strings );
    if ( max_len < l )
	 max_len = l;
  }
  return max_len+1;
}


/* Lookup a name string list from one array and return its index.
 * If not found, the last index (to NULL) is returned. */
static int
lookup ( SANE_String_Const name, SANE_String_Const const list[] )
{
  int i;
  for ( i=0; list[i] && strcmp (list[i], name); i++ )
    ;
  return i;
}


static int big2_to_int( SANE_Byte const * buffer )
{
  return buffer[0]*256 + buffer[1];
}

static int big3_to_int( SANE_Byte const * buffer )
{
  return big2_to_int(buffer)*256 + buffer[2];
}

static void int_to_big2( unsigned char * buffer, int value )
{
  buffer[0] = (unsigned char)( value >> 8 );
  buffer[1] = (unsigned char)value;
}

static void int_to_big3( unsigned char * buffer, int value )
{
  int_to_big2( buffer, value >> 8 );
  buffer[2] = (unsigned char)value;
}

static void int_to_big4( unsigned char * buffer, int value )
{
  int_to_big2( buffer, value >> 16 );
  int_to_big2( buffer+2, value );
}


static int fix_mm_to_pt ( SANE_Fixed mm, SANE_Int dpi) /*--------------------*/
{
  return (int)( SANE_UNFIX(mm) * dpi / 25.4 + .5 );
}


#define nullfree( p ) (void)( free( p ), p = NULL )


/******************************************************************************
 *
 * SANE option management
 */

/* temporarily outsourced, until VM6552 adaption is finished: */
#include "teco3_opt.c"


/******************************************************************************
 *
 * Scanner detection and management
 */


/* Park the CCD ============================================================ */

static SANE_Status
teco_reset_window ( Teco_Scanner * dev )
{
  unsigned char cdb[10] = { SCSI_OBJECT_POSITION };
  /* int_to_big3( cdb+2, position ); */

  return sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, NULL, NULL );
}


/* ========================================================================= */

void
sane_cancel ( SANE_Handle handle )
{
  Teco_Scanner *dev = handle;

  if ( dev->scanning )
    teco_reset_window( dev );

  dev->scanning = SANE_FALSE;
}


/* ========================================================================= */

void
sane_close ( SANE_Handle handle )
{
  Teco_Scanner *dev = handle;
  int i;

  sane_cancel( dev );
  if ( dev->sfd != -1 )
  {
    sanei_scsi_close( dev->sfd );
    dev->sfd = -1;
  }

  nullfree( dev->image );

  for ( i = 1; i < OPT_NUM_OPTIONS; i++ )
    if ( dev->opt[i].type == SANE_TYPE_STRING )
      nullfree( dev->val[i].s ); 
}


/* ========================================================================= */

void
sane_exit ( void )
{
  while( scanners )
  {
    Teco_Scanner * dev = scanners;
    scanners = dev->next;
    sane_close( dev );
    free( dev->devicename );
    free( dev );
  }
  nullfree( devlist );
}


/* ========================================================================= */

/* Inquire a device and check, if it is supported. ========================= */

static int
teco_identify_scanner ( Teco_Scanner * dev )
{
  SANE_Status status;

  unsigned char buffer[96];
  unsigned char cdb[6] = { SCSI_INQUIRY, 0,0,0, sizeof buffer };
  size_t size = sizeof buffer;
  SANE_String_Const const *scanner;

  status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, buffer, &size );
  if ( status )
  {
    DBG (DBG_error, "teco_identify_scanner: Inquiry of %s failed (status %s).\n",
				dev->devicename, sane_strstatus(status) );
    return SANE_FALSE;
  }

  hexdump( DBG_info2, "inquiry", buffer, size );

  if ( size < 68 || buffer[4] + 5 < 68 )
  {
    DBG( DBG_error, "teco_identify_scanner: Inquiry of %s returned %d bytes "
		    "(<68), or buffer[4]=%d (<0x43). This device is ignored.\n",
				dev->devicename, (int)size, buffer[4] );
    return SANE_FALSE;
  }

  dev->scsi_type = buffer[0] & 0x1f;
  memcpy( dev->scsi_teco_name, buffer + 0x2A, 0x0B );

  for ( scanner = scanner_list; *scanner; scanner++ )
  {
    if ( dev->scsi_type==6 && 0==strcmp(dev->scsi_teco_name, *scanner) )
    {
      memcpy( dev->scsi_vendor,  buffer + 0x08, 0x08 );
      memcpy( dev->scsi_product, buffer + 0x10, 0x10 );
      memcpy( dev->scsi_version, buffer + 0x20, 0x04 );

      dev->sane.name = dev->devicename;
      dev->sane.vendor = dev->scsi_vendor;
      dev->sane.model = dev->scsi_product;
      dev->sane.type = "flatbed scanner";	/* I18N? */

      DBG( DBG_error, "Supported device is \"%s\" \"%s\" \"%s\" \"%s\"\n",
				dev->scsi_vendor, dev->scsi_product,
				dev->scsi_version, dev->scsi_teco_name );

      dev->scsi_x_dpi.min = big2_to_int( buffer+54 );
      dev->scsi_x_dpi.max = big2_to_int( buffer+56 );

      dev->scsi_y_dpi.min = big2_to_int( buffer+58 );
      dev->scsi_y_dpi.max = big2_to_int( buffer+60 );

      dev->scsi_x_range = big2_to_int( buffer+62 );
      dev->scsi_y_range = big2_to_int( buffer+64 );

      dev->scsi_unit = big2_to_int( buffer+66 );
      if ( dev->scsi_unit < 0 )
      {
        /* guess unit by: x_range is 8.5": */
        dev->scsi_unit = dev->scsi_x_range * 2 / 17;
      }
      dev->x_range.max = SANE_FIX( dev->scsi_x_range * 25.4 / dev->scsi_unit );
      dev->y_range.max = SANE_FIX( dev->scsi_y_range * 25.4 / dev->scsi_unit );
      dev->dpi_range.max = MAX( dev->scsi_x_dpi.max, dev->scsi_y_dpi.max );

      DBG ( DBG_info, "resolution: X=%d..%ddpi, %d..%ddpi\n",
				dev->scsi_x_dpi.min, dev->scsi_x_dpi.max,
				dev->scsi_y_dpi.min, dev->scsi_y_dpi.max );
      DBG ( DBG_info, "scan area: %dx%d dots @ %ddpi\n",
			dev->scsi_x_range, dev->scsi_y_range, dev->scsi_unit );

     return SANE_TRUE;
    }
  }
  
  return SANE_FALSE;
}


/* ========================================================================= */


/* SCSI sense callback for SANE. These scanners never set asc or ascq. ===== */

static SANE_Status
teco_sense_handler ( int __sane_unused__ scsi_fd, unsigned char *result, void __sane_unused__ *dev )
{
  int sensekey = result[2] & 0x0f;
  int len = 7 + result[7];
  int errorcode = result[0] & 0x7f;

  DBG( DBG_proc, "teco_sense_handler: enter\n" );

  hexdump( DBG_info2, "sense", result, len );

  if ( errorcode != 0x70 )
  {
    DBG( DBG_error, "teco_sense_handler: invalid sense key error code (%d)\n",
								errorcode );
    return SANE_STATUS_IO_ERROR;
  }

  if ( len < 14 )
  {
    DBG( DBG_error, "teco_sense_handler: sense too short, no ASC/ASCQ\n" );
    return SANE_STATUS_IO_ERROR;
  }

  DBG( DBG_sense, "teco_sense_handler: sense=%d\n", sensekey );

  if (sensekey == 0x00)
    return SANE_STATUS_GOOD;

  return SANE_STATUS_IO_ERROR;
}


/* Attach a scanner. ======================================================= */
static SANE_Status
attach_scanner ( const char *devicename )
{
  Teco_Scanner *dev;
  SANE_Status status;

  /* Check if we know this device name. */
  for ( dev = scanners; dev; dev = dev->next )
  {
    if ( 0==strcmp(dev->devicename, devicename) )
    {
      DBG( DBG_info, "device is already attached\n" );
      return SANE_STATUS_GOOD;
    }
  }

  /* Allocate a new scanner entry. */
  dev = calloc( sizeof(Teco_Scanner), 1 );
  if ( dev == NULL )
  {
    DBG( DBG_error, "ERROR: not enough memory\n" );
    return SANE_STATUS_NO_MEM;
  }

  dev->sfd = -1;
  DBG (DBG_info, "attach_scanner: opening %s\n", devicename);

  status = sanei_scsi_open( devicename, &dev->sfd, teco_sense_handler, dev );
  if ( status )
  {
    DBG( DBG_error, "ERROR: attach_scanner: open failed (%s)\n",
						sane_strstatus(status) );
    free( dev );
    return SANE_STATUS_INVAL;
  }

  dev->devicename = strdup( devicename );
  /* Now, check that it is a scanner we support. */
  if ( !teco_identify_scanner( dev ) )
  {
    sanei_scsi_close( dev->sfd );
    free( dev->devicename );
    free( dev );
    return SANE_STATUS_INVAL;
  }

  sanei_scsi_close( dev->sfd );
  dev->sfd = -1;

  /* Link the scanner with the others. */
  dev->next = scanners;
  scanners = dev;

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_init ( SANE_Int * version_code, SANE_Auth_Callback __sane_unused__ authorize )
{
  DBG_INIT();

  DBG( DBG_error, "This is sane-teco3 version %d.%d-%d\n", V_MAJOR, V_MINOR, BUILD );
  DBG( DBG_error, "Copyright (C) 2002 by Frank Zago\n" );
  DBG( DBG_error, "Copyright (C) 2008 by Ralph D. Ungermann\n" );

  if ( version_code )
    *version_code = SANE_VERSION_CODE( V_MAJOR, V_MINOR, BUILD );

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_get_devices ( const SANE_Device *** device_list, SANE_Bool __sane_unused__ local_only )
{
  Teco_Scanner *dev;
  int n_dev = 0;
  FILE *fp;

  fp = sanei_config_open( TECO_CONFIG_FILE );
  if ( !fp )
  {
    /* default to /dev/scanner instead of insisting on config file */
    attach_scanner( "/dev/scanner" );
  }
  else
  {
    char dev_name[PATH_MAX];
    while ( sanei_config_read(dev_name, sizeof dev_name, fp) )
    {
      if ( '\0'==dev_name[0] || '#'==dev_name[0] )
	continue;	/* ignore comments(#) and empty lines(\0) */

      sanei_config_attach_matching_devices( dev_name, attach_scanner );
    }
    fclose (fp);
  }

  for ( dev=scanners; dev; dev = dev->next )
    ++n_dev;

  *device_list = NULL;
  free( devlist );
  devlist = calloc( sizeof devlist[0], n_dev+1 );
  if ( !devlist )
    return SANE_STATUS_NO_MEM;

  for ( dev = scanners; dev; dev = dev->next )
    devlist[--n_dev] = &dev->sane;

  *device_list = devlist;

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_open( SANE_String_Const devicename, SANE_Handle * handle )
{
  SANE_Status status;
  Teco_Scanner *dev;
  int i;

  /* search for devicename */
  if ( devicename[0] )
  {
    DBG( DBG_info, "sane_open: devicename=%s\n", devicename );

    for ( dev = scanners; dev; dev = dev->next )
    {
      if ( 0==strcmp(dev->sane.name, devicename) )
	break;
    }
    if ( dev==NULL )
    {
      status = attach_scanner( devicename );
      if ( status )
	return status;
      dev = scanners;
      if ( dev==NULL || strcmp(dev->devicename, devicename) )
      {
	DBG( DBG_error, "Internal error: scanner '%s' opened, but '%s' found.\n",
			devicename, dev ? dev->devicename : "nothing" );
	return SANE_STATUS_INVAL;
      }
    }
  }
  else
  {
    DBG( DBG_sane_info, "sane_open: no devicename, opening first device\n" );

    if ( scanners==NULL )
    {
      const SANE_Device **device_list = NULL;
      status = sane_get_devices( &device_list, 1 );
      if ( status )
	return status;
    }
    dev = scanners;
  }

  if ( dev==NULL )
  {
    DBG( DBG_error, "Scanner '%s' not found.\n", devicename );
    return SANE_STATUS_INVAL;
  }

  status = sanei_scsi_open( devicename, &dev->sfd, teco_sense_handler, dev );
  if ( status )
  {
    DBG( DBG_error, "ERROR: sane_open: SCSI open failed (%s)\n",
						sane_strstatus(status) );
    return SANE_STATUS_INVAL;
  }

  teco_init_options(dev);

  /* Initialize the gamma table. */
  for (i = 0; i < GAMMA_LENGTH; i++)
    {
      dev->gamma_R[i] = i / 4;
      dev->gamma_G[i] = i / 4;
      dev->gamma_B[i] = i / 4;
      dev->gamma_GRAY[i] = i / 4;
    }

  *handle = dev;

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_set_io_mode ( SANE_Handle handle, SANE_Bool non_blocking )
{
  Teco_Scanner *dev = handle;

  if ( !dev->scanning )
    return SANE_STATUS_INVAL;

  if ( non_blocking )
    return SANE_STATUS_UNSUPPORTED;

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_get_select_fd (SANE_Handle __sane_unused__ handle, SANE_Int __sane_unused__ * fd)
{
  DBG( DBG_proc, "sane_get_select_fd: not implemented.\n" );
  return SANE_STATUS_UNSUPPORTED;
}


/******************************************************************************
 *
 * Scanner Operation
 */


/* Set a window. =========================================================== */
static SANE_Status
teco_set_window ( Teco_Scanner * dev )
{
  unsigned char buffer[69] = { 0 };
  unsigned char cdb[10] = { SCSI_SET_WINDOW, 0,0,0,0,0, 0,0,sizeof buffer };

  /* size of the windows descriptor block */
  buffer[7] = (unsigned char)( sizeof buffer - 8 );

  if ( dev->val[OPT_PREVIEW].w )
  {
    int_to_big2( buffer+10, 50 );
    int_to_big2( buffer+12, 50 );
    int_to_big4( buffer+22, dev->scsi_x_range );
    int_to_big4( buffer+26, dev->scsi_y_range );
  }
  else
  {
    int_to_big2( buffer+10, MIN( dev->scsi_x_dpi.max, dev->val[OPT_RESOLUTION].w) );
    int_to_big2( buffer+12, MIN( dev->scsi_y_dpi.max, dev->val[OPT_RESOLUTION].w) );
    int_to_big4( buffer+14, fix_mm_to_pt(dev->val[OPT_TL_X].w, dev->scsi_unit));
    int_to_big4( buffer+18, fix_mm_to_pt(dev->val[OPT_TL_Y].w, dev->scsi_unit));
    int_to_big4( buffer+22, fix_mm_to_pt(dev->val[OPT_BR_X].w - dev->val[OPT_TL_X].w, dev->scsi_unit));
    int_to_big4( buffer+26, fix_mm_to_pt(dev->val[OPT_BR_Y].w - dev->val[OPT_TL_Y].w, dev->scsi_unit));
  }

  buffer[31] = dev->val[OPT_THRESHOLD].w;
  buffer[33] = mode_4scsi[dev->scan_mode];
  buffer[34] = mode_depth[dev->scan_mode];
  buffer[36] = lookup( dev->val[OPT_DITHER].s, dither_list );

  /* Unknown - invariants */
  buffer[37] = 0x80;	/* ? RIF (Reverse Image Format) */

  /* model specific? */
  buffer[48] = 0x01;	/* ? Color Channel: 0=R 1=G 2=B (BW and Grey only) */
  buffer[50] = 0x02;	/* ? Calibration: 0=yes 2=no */
  buffer[53] = 0xff;	/* ? expected line size */
  buffer[57] = 0xff;	/* ? expected number of lines */
  buffer[61] = 0xff;
  buffer[65] = 0xff;

  hexdump( DBG_info2, "set scan window", buffer, sizeof buffer );

  return sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb,
					buffer, sizeof buffer, NULL, NULL );
}


/* - info==NULL:
 *   set dev->params to the exact values retrieved from the scanner.
 * - with info:
 *   return the number of bytes waiting and the color type of the most recently
 *   read line: 0=R, 1=G, 2=B, 3=n/a (BW, gray or RGB triples).
 */
static SANE_Status
buffer_status ( Teco_Scanner * dev, size_t info[2] /* { size, color } */ )
{
  SANE_Status status;

  unsigned char buffer[18] = { 0 };
  size_t size = sizeof buffer;
  unsigned char cdb[10] = { SCSI_GET_DATA_BUFFER_STATUS, 1/*wait*/, 0,0,0,0,0, 0,sizeof buffer };

  if ( !dev->scanning )
    return SANE_STATUS_CANCELLED;

  status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, buffer, &size );
  if (status)
    return status;

  if ( size < 18 )
  {
    DBG( DBG_error, "buffer_status: recieved only %d bytes instead of 18\n",
						(int)size );
    return SANE_STATUS_IO_ERROR;
  }
  hexdump( DBG_info2, "buffer_status recieved", buffer, size );

  if ( info ) /* scan in progress: report current buffer state. */
  {
    info[0] = big3_to_int( buffer+9 );

    if ( dev->scan_mode==TECO_COLOR && buffer[17]==0x07 )
    {
      /* If there is not enough scanner RAM, the colors are batched. */
      info[1] = ( buffer[16]-1 ) & 3;
      DBG( DBG_info, "buffer_status: data: %ld, color mode=%d-%d\n", 
					(long)info[0], buffer[16], buffer[17] );
    }
    else
    {
      info[1] = 3; /* data is suitable for sane */
    }
  }
  else /* Beginning of a scan: fill params with exact values. */
  {
    dev->params.lines = big2_to_int( buffer+12 );
    dev->params.last_frame = SANE_TRUE;
    dev->params.format = mode_frame[dev->scan_mode];

    dev->params.depth = mode_depth[dev->scan_mode];
    dev->params.bytes_per_line = big2_to_int( buffer+14 ) * mode_color[dev->scan_mode];
    dev->params.pixels_per_line = dev->params.bytes_per_line
				* 8 / dev->params.depth / mode_color[dev->scan_mode];

    DBG( DBG_info, "The Scanner has %dKB RAM\n", big3_to_int(buffer+6)/1024 );
    DBG( DBG_info, "Next image has %d lines, %d bytes/line, %d %dbit colors.\n",
				dev->params.lines, dev->params.bytes_per_line,
				mode_color[dev->scan_mode], dev->params.depth );
  }

  return status;
}


/* ========================================================================= */


SANE_Status
sane_get_parameters ( SANE_Handle handle, SANE_Parameters * params )
{
  Teco_Scanner *dev = handle;

  if ( !dev->scanning ) /* dev->params is not yet set; so estimate */
  {
    dev->params.depth = mode_depth[dev->scan_mode];
    dev->params.last_frame = SANE_TRUE;
    dev->params.format = mode_frame[dev->scan_mode];

    dev->params.lines = 
		  fix_mm_to_pt( dev->val[OPT_BR_Y].w - dev->val[OPT_TL_Y].w, 
			MIN(dev->scsi_y_dpi.max, dev->val[OPT_RESOLUTION].w) );

    dev->params.pixels_per_line =
		  fix_mm_to_pt( dev->val[OPT_BR_X].w - dev->val[OPT_TL_X].w, 
			MIN(dev->scsi_x_dpi.max, dev->val[OPT_RESOLUTION].w) );

    dev->params.bytes_per_line  = dev->params.pixels_per_line
				* dev->params.depth * mode_color[dev->scan_mode] / 8;
  }

  /* Return the current values. */
  if ( params )
    *params = dev->params;

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


static void words_to_bytes( unsigned char byte[], SANE_Word const word[], size_t n )
{
  while ( n-- )
    *byte++ = (unsigned char) *word++;
}


/* Send the gamma ---------------------------------------------------------- */

static SANE_Status
teco_send_gamma ( Teco_Scanner * dev )
{
  unsigned char gamma[4*GAMMA_LENGTH] = { 0 }; /* R, (G|gray), B, (unused) */
  /* Q: why 4*, if only 3 are actually used? */

  unsigned char cdb[10] = { SCSI_SEND_10, 0,3 /*dtc*/, 0,0,2 /*dtq*/ };

  int_to_big3( cdb+6, sizeof gamma );

  if ( dev->val[OPT_CUSTOM_GAMMA].w )
  {
    if ( dev->scan_mode == TECO_GRAYSCALE )
    {
      words_to_bytes( gamma+1*GAMMA_LENGTH, dev->gamma_GRAY, GAMMA_LENGTH );
    }
    else
    {
      words_to_bytes( gamma+0*GAMMA_LENGTH, dev->gamma_R, GAMMA_LENGTH );
      words_to_bytes( gamma+1*GAMMA_LENGTH, dev->gamma_G, GAMMA_LENGTH );
      words_to_bytes( gamma+2*GAMMA_LENGTH, dev->gamma_B, GAMMA_LENGTH );
    }
  }
  else
  {
    int i;
    for (i = 0; i < GAMMA_LENGTH; i++)
      gamma[i] = gamma[GAMMA_LENGTH+i] = gamma[2*GAMMA_LENGTH+i] = i*256/GAMMA_LENGTH;
  }

  return sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, gamma, sizeof gamma, NULL, NULL );
}


SANE_Status
sane_start ( SANE_Handle handle )
{
  SANE_Status status = SANE_STATUS_GOOD;
  Teco_Scanner *dev = handle;

  if ( dev->scanning )
    return SANE_STATUS_DEVICE_BUSY;

#ifdef TECO3_TESTCODE
  {
    FILE *fp = sanei_config_open( TECO_CONFIG_FILE );
    if ( fp )
    {
      char buffer[128] = "";
      while ( sanei_config_read( buffer, sizeof buffer, fp ) )
      {
        if ( 0==strncmp(buffer, "#TEST=", 6) )
        {
	  dev->test_mode = lookup( buffer+6, test_mode_list );
	  DBG( DBG_error, "setting TEST=%s\n", test_mode_list[dev->test_mode] );
	}
      }
      fclose( fp );
    }
  }
#endif

  /* Wait for scanner ready. */
  {
    int timeout = 0;
    unsigned char cdb[6] = { SCSI_TEST_UNIT_READY };
    do
    {
      status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, NULL, NULL );
      if ( status )
	sleep(1);

      hexdump( DBG_info2, "SCSI_TEST_UNIT_READY", cdb, sizeof cdb );

      if (++timeout==30 ) /* seconds */
      {
	DBG( DBG_error, "sane_start: scanner not ready -- timeout!\n" );
	return SANE_STATUS_IO_ERROR;
      }
    } while ( status );
  }

  /* query sense -- whatever this is */
  {
    unsigned char buffer[256];
    unsigned char cdb[6] = { SCSI_REQUEST_SENSE, 0,0,0, 255 /*buflen*/ };
    size_t size = sizeof buffer;

    status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, buffer, &size );
    hexdump( DBG_info2, "SCSI_REQUEST_SENSE", buffer, MIN( buffer[7]+8u, size ) );
    if (status)
      return status;
  }

  dev->scanning = SANE_TRUE;

  teco_reset_window( dev );

  status = teco_set_window( dev );
  if (status)
    return status;

  /* Now set dev->params correct: */
  status = buffer_status( dev, NULL );
  if (status)
    return status;

  /* Compute the length necessary in image. */
  dev->image_size = dev->params.lines * dev->params.bytes_per_line;

  /* We need +line/3 as a raster buffer, min. 0x7800 for calibration: */
  dev->image = dev->image_head = dev->image_tail = 
	malloc( MAX( dev->image_size + dev->params.bytes_per_line/3, 0x7800 ) );

  if ( dev->image == NULL )
    return SANE_STATUS_NO_MEM;

  dev->image_stop = dev->image + dev->image_size; /* raster buffer follows */
  dev->image_line[0] = dev->image + 0; /* first RED pixel */
  dev->image_line[1] = dev->image + 1;
  dev->image_line[2] = dev->image + 2;

  /* calibrate scanner */
  {
    size_t size = MIN( dev->image_size, 0x7800 /* 12lines, 8.5inch, 300dpi? */ );
    unsigned char cdb[6] = { SCSI_VENDOR_09 };
    int_to_big2( cdb+3, size );
    status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, dev->image, &size );

    hexdump( DBG_info2, "calibration", dev->image, MIN(size,256) );
    /* shouldn't we do something with these bytes? */

    if (status)
      return status;
  }

  memset( dev->image, 0xff, dev->image_size ); /* blank to ease debugging... */

  /* Do some vendor specific stuff. */
  {
    unsigned char cdb[6] = { SCSI_VENDOR_0E };
    status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, NULL, NULL );
    if ( status )
      return status;
  }

  status = teco_send_gamma( dev );
  if ( status )
    return status;

  /* Start a scan. */
  {
    unsigned char cdb[6] = { SCSI_SCAN };
    status = sanei_scsi_cmd2( dev->sfd, cdb, sizeof cdb, NULL, 0, NULL, NULL );
    if ( status )
      return status;
  }

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


static SANE_Status
scsi_read ( int sfd, SANE_Byte *buffer, size_t *size )
{
  SANE_Status status;
  unsigned char cdb[10] = { SCSI_READ_10 };
  int_to_big3( cdb+6, *size );

  status = sanei_scsi_cmd2( sfd, cdb, sizeof cdb, NULL, 0, buffer, size );
  if ( status )
    DBG (DBG_error, "read error: %s.\n", sane_strstatus(status) );
  return status;
}


/* Wait for and read available data from the scanner to the image buffer.
 * On success, image_tail is moved to the end of the data (usually at the
 * start of line, because the scanner alsways delivers whole lines).
 */
static SANE_Status
teco_read ( Teco_Scanner * dev )
{
  SANE_Status status;
  size_t info[2] = { 0, 0 };	/* #data , color */
  int timeout = 0;

  if ( dev->image_tail >= dev->image_stop )
    return SANE_STATUS_EOF;
  
  for ( ;; )	/* Wait for data */
  {
    status = buffer_status( dev, info ); /* also checks if cancelled */
    if ( status )
      return status;

    if ( info[0] )	/* data available */
      break;
    usleep( 100000 );	/* 0.1 seconds */

    if ( ++timeout==300 )
    {
      DBG( DBG_error, "sane_read: no more data -- timeout!\n" );
      return SANE_STATUS_EOF;
    }
  }

  if ( info[1]>2 /*correct data order*/ || dev->test_mode==2 /*bulk*/ )
  {
    /* Scan data arrives in the right order. We may append it all 
     * right behind image_tail.
     */
    info[0] = MIN( info[0], (size_t)( dev->image_stop-dev->image_tail ) );

    DBG( DBG_info, "reading %d bytes (%d lines)\n",
			(int)info[0], (int)( info[0]/dev->params.bytes_per_line ) );
    status = scsi_read( dev->sfd, dev->image_tail, &info[0] );
    if ( status )
      return status;

    if ( dev->scan_mode==TECO_BW )
    {
      /* The bits are inverted and mirrored: 11010001 --> 01110100.
       * We fix that via a 4bit lookup table: hi.lo --> mirror[lo].mirror[hi]
       */
      static unsigned const char mirror[16] = 
			{ 0xF, 0x7, 0xB, 0x3, 0xD, 0x5, 0x9, 0x1,
			  0xE, 0x6, 0xA, 0x2, 0xC, 0x4, 0x8, 0x0 };
      size_t i;
      for ( i = 0; i < info[0]; i++ )
	dev->image_tail[i] = mirror[ dev->image_tail[i]%16 ]*16
			   + mirror[ dev->image_tail[i]/16%16 ];
    }

#ifdef TECO3_TESTCODE
    if ( dev->test_mode==2 )
    {
      size_t ofs = 2 - ( (dev->image_tail-dev->image) % dev->params.bytes_per_line + 2 ) % 3;
      if ( ofs+2 < info[0] ) /* enough space to write color as first pixel: */
      {
	dev->image_tail[ofs+0] = -( info[1]==0 );
	dev->image_tail[ofs+1] = -( info[1]==1 );
	dev->image_tail[ofs+2] = -( info[1]==2 );
      }
    }
#endif

    dev->image_tail += info[0];
  }
  else /* info[1]={0,1,2} <-- old value from the previous read() */
  {
    /* Scan data arrives in line batches per color. We read it line by line, 
     * as we must query the corresponding color of each batch.
     * Data is appended at image_line[color]+0,3,6,9...
     * Finally, image_tail advances to the minimum of all three colors.
     */
    do /* we'll flush the buffer, but won't wait for more */
    {
      size_t raster = dev->params.bytes_per_line/3;
      static char const color_id[4] = { 'R', 'G', 'B', 'X' };

      status = scsi_read( dev->sfd, dev->image_stop, &raster );
      if ( status )
	return status;

      if ( raster*3 != (size_t)dev->params.bytes_per_line )
	return SANE_STATUS_IO_ERROR;

      status = buffer_status( dev, info ); /* sets info[1] to current color */
      if ( status )
        return status;
      if ( info[1] > 2 )
        return SANE_STATUS_IO_ERROR;
        
      DBG( DBG_info2, "next lines are: R=%d G=%d B=%d\n",
		(int)( dev->image_line[0]-dev->image )/dev->params.bytes_per_line,
		(int)( dev->image_line[1]-dev->image )/dev->params.bytes_per_line,
		(int)( dev->image_line[2]-dev->image )/dev->params.bytes_per_line );
      DBG( DBG_info, "reading %d bytes of color %c\n",
					(int)raster, color_id[info[1]] );

      if ( dev->image_line[info[1]] >= dev->image_stop )
      {
	/* Something went wrong. But we must account these bytes anywhere,
	 * s.th. dev->image_tail will finally grow behind the last row.
	 * We might also return an error here.
	 */
	DBG( DBG_error, "ERROR: %c frame is already complete! Data dropped!\n"
			"  ****************************\n", color_id[info[1]] );

	if ( dev->image_line[0] <= dev->image_stop )
	  info[1] = 0;
	if ( dev->image_line[1] <= dev->image_stop )
	  info[1] = 1;
	if ( dev->image_line[2] <= dev->image_stop )
	  info[1] = 2;
      }
      else
      {
#ifdef TECO3_TESTCODE
	if ( dev->test_mode==1 ) /* RRRR GGGG BBBB instead of RGB RGB RGB RGB */
	{
	  SANE_Byte *dst = dev->image_line[info[1]] + info[1]*(raster-1);
	  size_t ofs = 2 - ( (dst-dev->image) % dev->params.bytes_per_line + 2 ) % 3;

	  memcpy( dst, dev->image_stop, raster );

	  if ( ofs+18 < raster ) /* enough space for raster number as binary */
	  {
	    int id = ( dev->image_line[0]-dev->image )/dev->params.bytes_per_line
		   + ( dev->image_line[1]-dev->image )/dev->params.bytes_per_line
		   + ( dev->image_line[2]-dev->image )/dev->params.bytes_per_line;
	    size_t i;
	    for ( i = 6; i > 0; i--, id /= 2 )
	      dst[ofs+3*i-3] = dst[ofs+3*i-2] = dst[ofs+3*i-1] = ( id & 0x1 )-1;
	  }
        }
	else if ( dev->test_mode==3 /* single rasters with color dot: */ )
	{
	  SANE_Byte *dst = dev->image_line[0];
	  size_t ofs = 2 - ( (dst-dev->image) % dev->params.bytes_per_line + 2 ) % 3;

	  memcpy( dst, dev->image_stop, raster );

	  if ( ofs+2 < raster ) /* write color (e.g. 00FF00) at first pixel: */
	  {
	    dst[ofs+0] = -( info[1]==0 );
	    dst[ofs+1] = -( info[1]==1 );
	    dst[ofs+2] = -( info[1]==2 );
	  }
	  dev->image_line[0] += raster;
	  dev->image_tail = dev->image + ( dev->image_line[0]-dev->image )
		      / dev->params.bytes_per_line * dev->params.bytes_per_line;
	  return SANE_STATUS_GOOD; /* avoid that "advance image_tail" below */
	}
	else /* normal operation */
#endif
	{
	  size_t i;
	  for (i = 0; i < raster; i++)
	    dev->image_line[info[1]][3*i] = dev->image_stop[i];
	}
      }

      /* color info[1] has just been filled: */
      dev->image_line[info[1]] += dev->params.bytes_per_line;

    } while ( info[0] ); /* repeat, if more data is available */

    /* advance image_tail to smallest of all color lines: */
    dev->image_tail = MIN( dev->image_line[0], dev->image_line[1]-1 );
    dev->image_tail = MIN( dev->image_tail,    dev->image_line[2]-2 );
  }

  return SANE_STATUS_GOOD;
}


/* ========================================================================= */


SANE_Status
sane_read ( SANE_Handle handle, SANE_Byte * buf, SANE_Int max_len, SANE_Int * len )
{
  Teco_Scanner *dev = handle;

  *len = 0;

  if ( !dev->scanning )
    return SANE_STATUS_CANCELLED;

  if ( dev->image_head >= dev->image_stop )
    return SANE_STATUS_EOF;

  while ( dev->image_tail <= dev->image_head  )
  {
    SANE_Status status = teco_read(dev); /* may be cancelled or time-out */
    if ( status )
      return status;
  }

  *len = MIN( max_len, (SANE_Int)( dev->image_tail - dev->image_head ) );
  memcpy( buf, dev->image_head, *len );

  dev->image_head += *len;

  DBG (DBG_info, "sane_read: leave, %d lines left\n",
	   (int)( dev->image_stop-dev->image_head )/dev->params.bytes_per_line );

  return SANE_STATUS_GOOD;
}
