/* sane - Scanner Access Now Easy.

   Copyright (C) 2002 Frank Zago (sane at zago dot net)
   Major extensions (VM6552)
   Copyright (C) 2008 Ralph D. Ungermann

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice. 
*/

/* 
	$Id: teco3.h,v 1.5 2005-07-07 11:55:42 fzago-guest Exp $
*/

/* Commands supported by the scanner. */
#define SCSI_TEST_UNIT_READY		0x00
#define SCSI_REQUEST_SENSE		0x03
#define SCSI_VENDOR_09			0x09
#define SCSI_VENDOR_0E			0x0E
#define SCSI_INQUIRY			0x12
#define SCSI_MODE_SELECT		0x15
#define SCSI_SCAN			0x1b
#define SCSI_SET_WINDOW			0x24
#define SCSI_SEND_10			0x2a
#define SCSI_READ_10			0x28
#define SCSI_OBJECT_POSITION		0x31
#define SCSI_GET_DATA_BUFFER_STATUS	0x34


/*--------------------------------------------------------------------------*/

#define GAMMA_LENGTH 0x400	/* number of value per color */

/*--------------------------------------------------------------------------*/

enum Teco_Option
{
  OPT_NUM_OPTS = 0,

  OPT_MODE_GROUP,
  OPT_MODE,			/* scanner modes */
  OPT_RESOLUTION,		/* X and Y resolution */

  OPT_GEOMETRY_GROUP,
  OPT_TL_X,			/* upper left X */
  OPT_TL_Y,			/* upper left Y */
  OPT_BR_X,			/* bottom right X */
  OPT_BR_Y,			/* bottom right Y */

  OPT_ENHANCEMENT_GROUP,
  OPT_CUSTOM_GAMMA,		/* Use the custom gamma tables */
  OPT_GAMMA_VECTOR_R,		/* Custom Red gamma table */
  OPT_GAMMA_VECTOR_G,		/* Custom Green Gamma table */
  OPT_GAMMA_VECTOR_B,		/* Custom Blue Gamma table */
  OPT_GAMMA_VECTOR_GRAY,	/* Custom Grayscale Gamma table */
  OPT_THRESHOLD,		/* Threshold */
  OPT_DITHER,
  OPT_PREVIEW,

  /* must come last: */
  OPT_NUM_OPTIONS
};

/*--------------------------------------------------------------------------*/


/* aren't there better predefined strings? */
#define BLACK_WHITE_STR		SANE_I18N("Black & White")
#define GRAY_STR		SANE_I18N("Grayscale")
#define COLOR_STR		SANE_I18N("Color")

enum Teco_Mode
{
  TECO_BW,
  TECO_GRAYSCALE,
  TECO_COLOR,
  TECO_N_MODES
};


/*--------------------------------------------------------------------------*/

/* Define a scanner occurence. */
typedef struct Teco_Scanner
{
  struct Teco_Scanner *next;
  SANE_Device sane;

  char *devicename;		/* malloc()'d */
  int sfd;			/* device handle */

  /* Infos from inquiry. */
  char scsi_type;
  char scsi_vendor[9];
  char scsi_product[17];
  char scsi_version[5];
  char scsi_teco_name[12];	/* real name of the scanner */

  SANE_Range scsi_x_dpi;	/* x-dpi range */
  SANE_Range scsi_y_dpi;	/* x-dpi range */
  SANE_Int   scsi_x_range;	/* max width */
  SANE_Int   scsi_y_range;	/* max height */
  SANE_Int   scsi_unit;		/* dpi used for set_window */

  SANE_Range dpi_range;		/* max dpi */
  SANE_Range x_range;		/* max image width [mm] */
  SANE_Range y_range;		/* max image height [mm] */

  /* Scanning handling. */
  volatile int scanning;	/* TRUE if a scan is running.
				   May be cleared on cancel. */
  enum Teco_Mode scan_mode;

#ifdef TECO3_TESTCODE
  int test_mode;		/* set in sane_scan() from config file */
#endif

  size_t image_size;		/* size of image */
  SANE_Byte *image;		/* keep the raw image here */
  SANE_Byte *image_stop;	/* raster buffer behind image */
  SANE_Byte *image_head;	/* first unread byte in image */
  SANE_Byte *image_tail;	/* first unscanned byte in image */
  SANE_Byte *image_line[3];	/* next unscanned pixel [RGB] */

  SANE_Parameters params;

  /* Options */
  SANE_Option_Descriptor opt[OPT_NUM_OPTIONS];
  Option_Value val[OPT_NUM_OPTIONS];

  /* Gamma table. 1 array per colour. */
  SANE_Word gamma_GRAY[GAMMA_LENGTH];
  SANE_Word gamma_R[GAMMA_LENGTH];
  SANE_Word gamma_G[GAMMA_LENGTH];
  SANE_Word gamma_B[GAMMA_LENGTH];
}
Teco_Scanner;

/*--------------------------------------------------------------------------*/

/* Debug levels. 
 * Should be common to all backends. */

#define DBG_error0  0
#define DBG_error   1
#define DBG_sense   2
#define DBG_warning 3
#define DBG_inquiry 4
#define DBG_info    5
#define DBG_info2   6
#define DBG_proc    7
#define DBG_read    8
#define DBG_sane_init   10
#define DBG_sane_proc   11
#define DBG_sane_info   12
#define DBG_sane_option 13
